# -*- coding: utf-8 -*-

# mapWOC
#
# Description: Process to poll results from nodes.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
#
# Copyright:
# Copyright (C) 2011 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import multiprocessing
from time import sleep
from Queue import Empty
import tempfile
from os import listdir, path, walk, rmdir
from os.path import isdir
import glob
import logging
import multiprocessing
import subprocess
import tarfile
import datetime

from sqlalchemy import *

from mapwoc.lib.result import Result, ResultType, Visit
from mapwoc.lib.url import URL
import mapwoc.lib.model.result
import mapwoc.lib.model.url
from mapwocgui.model.meta import Session

logger = logging.getLogger('master')

class ResultPollProcess(multiprocessing.Process):
    """Process that polls for results (rsnyc) of nodes.
       In case that any results where fetched, these are registered
       in the database.
    """

    SLEEP_STEP = 60

    def __init__(self, working_dir, nodes, result_queue, credentials_path, db):
        """Constructor, set up fields to sane values.

           working_dir Directory in which to create temporary dir. for syncing.
           nodes NodeConnection list of nodes with which to sync.
           result_queue Queue in which to write when results were found.
           credentials_path Path to certificates for ssh/rsync
           db Path to database in which to store the results.
        """

        multiprocessing.Process.__init__(self)
        self.__working_dir = working_dir
        self.__nodes = nodes
        self.result_queue = result_queue
        self.credentials_path = credentials_path
        self.db_file = db

    def __bind_to_db(self):
        """Bind the Session to database."""

        # TODO Find out whether sqlalchemy can protect from concurrency issues,
        #      and how.
        engine = create_engine('sqlite:///%s' % self.db_file) 
        c = engine.connect() 
        Session.configure(bind=engine) 
        

    def run(self):
        """Called upon call to start().
           Starts a Rsync process peridically.
        """

        self.__bind_to_db()

        while True:
            rsnc_subprocess = None

            # Call that rsync subprocess, sync from there into:

            for nodekey, node in self.__nodes.items():
                tempd = path.abspath(tempfile.mkdtemp(dir=self.__working_dir,
                                             prefix='resultsync_%s_' % nodekey))
                if not node.data_path:
                    logger.error("Node %s does not have a resultdir set. Is it running?", nodekey)
                    continue
                cmd = ['rsync', '-az',
                        '-e', ('"/usr/bin/ssh -oKbdInteractiveAuthentication=no '
                            '-oPasswordAuthentication=no '
                            '-oNumberOfPasswordPrompts=0 -i %s"'
                        ) % path.abspath(self.credentials_path),
                        '--remove-source-files',
                        '%s:%s' % (node.rsync_address, node.data_path),
                        tempd ]
                p = subprocess.Popen(" ".join(cmd),
                                stdin=subprocess.PIPE,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                shell=True)
                p.wait()

                if p.returncode != None:
                    if p.returncode == 0:
                        logger.info("Result Sync with node %s successful."
                                    % nodekey)
                        # Now, check what came and import it to db.
                    else:
                       stdout_value, stderr_value = p.communicate()
                       logger.warning("Could not sync results with node %s."
                                      % nodekey)
                       logger.debug("stdout: %s" % stdout_value)
                       logger.debug("stderr: %s" % stderr_value)
                       logger.debug("exit code: %d" % p.returncode)
                       logger.debug("command was: %s" % cmd)
                else:
                    logger.warning("Rsync without returncode.")

                # Walk over rsynced once, put in db/remove. Signal the master if something serious is happening.
                if self.__resultdir_to_db(tempd):
                    self.result_queue.put('marker')
                #(self.master.update_results)

            sleep(self.SLEEP_STEP)

    def __resultdir_to_db(self, resultdir):
        """Analyzes a result directory and imports the results into the db"""
        for dirpath, dirnames, filenames in walk(resultdir):
            if not filenames and not dirnames:
                # clean up empty directories also removes an empty
                # resultdir
                rmdir(dirpath)
            for filename in filenames:
                if not filename.endswith(".tar.gz"):
                    logger.warning("Unkown file %s in result dir %s" % \
                            (filename, dirpath))
                else:
                    if not self.__result_to_db(path.join(dirpath, filename)):
                        logger.warning("Could not extract result from %s" % filename)

    def __result_to_db(self, resultb):
        """Imports a result archive into the db.
           This function is a working prototype.
           parameter resultb should point to an existing tarball of results
        """

        # TODO Define what information is interesting, how it can be stored
        #      and how it can be found during processing of the dump.
        logger.debug("Entering result %s into database" % resultb)
        found = False
        try:
            tgz = tarfile.open(resultb)
        except tarfile.ReadError:
            logger.warning("Garbage in result directory (expected tgz): %s" % resultb)
            return False
        except IOError:
            logger.warning("Error accessing content of result directory: %s." % resultb)
            return False

        if 'result/Log/404.log' in tgz.getnames():
            # At least one 404.
            self.__enter_404(tgz.extractfile('result/Log/404.log'))
        else:
            # Over-simplify for testing purposes.
            if not 'result/Log/header.log' in tgz.getnames():
                #TODO
                logger.debug('Error in analysis or no traffic found. Report anyway.')
                return True

        if 'result/Log/header.log' in tgz.getnames():
            hl = tgz.extractfile('result/Log/header.log')
            logger.debug('Have a header log')
            lines = hl.readlines()
            if len(lines) <= 1:
                #TODO
                logger.debug('Have an empty header log')
                return True
            for line in lines[1:]:
                recs = line.strip().split()
                # fields in line: src-ip,time,dst(host),agent(crippled)
                if recs[2].startswith('10.0.0.1'):
                    logger.debug("Found the redirector query %s." % recs[2])
                    continue

                logger.debug("Making hard assumptions (http) about analysed url visit.")
                url_string = "http://%s" % recs[2]
                url = Session.query(URL).filter(URL.url == url_string).first()
                if url is None:
                    url = URL(url_string)
                    Session.add(url)
                t204 = Session.query(ResultType).filter(
                                    ResultType.type_name == 'RegularTraffic').first()
                if t204 is None:
                    t204 = ResultType('RegularTraffic', 204)
                result = Result(url, result_type=t204,
                                method='mapwocanalyse.pl/header.log/0.0.1')
                result.text = '- found in Log/header.log -'
                visit = Visit('unknown')
                visit.results.append(result)
                visit.dump_location='not decided'
                Session.add(result)
                Session.add(visit)
                Session.commit()
                logger.debug("Committed 204 result to database.")
                #found true?
                found = True
            #TODO its an error if there is no lines.
            #    resulttype = ResultType('NoTraffic', 204)
            #else:
            #    if len(hl.readlines()) > 2:
            #        resulttype = ResultType('RegularVisit', 200)
            #    else:
            #        resulttype = ResultType('InsufficientTraffic', 204)
            # could find google analyse in Log/Analyse.log
        return found

    def __enter_404(self, log):
        """ Enters a 404 result into the database """

        t400 = Session.query(ResultType).filter(
                         ResultType.type_name == '404').first()
        if t400 is None:
            t400 = ResultType('404', 404)

        for line in log.readlines():
            url_string = line.split('\t')[-1].strip()
            url = Session.query(URL).filter(
                                URL.url == url_string).first()
            if url is None:
                url = URL(url_string)
                Session.add(url)
            result = Result(url, result_type=t400,
                            method='mapwocanalyse.pl/404/0.0.1')
            result.text = line.strip()
            visit = Visit('unknown')
            visit.results.append(result)
            visit.dump_location='not decided'
            Session.add(result)
            Session.add(visit)
            Session.commit()
            logger.debug("Committed 404 result to database.")
