# -*- coding: utf-8 -*-

# mapWOC
#
# Description: KVM-Process related classes.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
#
# Copyright:
# Copyright (C) 2011 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import multiprocessing
import subprocess
from os import path, remove, chmod, makedirs
import stat
import socket
import tempfile
import logging
from Queue import Empty

import time
from time import sleep
from shutil import copy

from mapwoc.lib.hc import HC


logger = logging.getLogger("node")

# TODO Refactor KVMHCRunProcess to inherit from TimedProcess which in turn
#      can be used where other processes are used to e.g. query node state.
class TimedProcess(multiprocessing.Process):

    def __init__(self):
        multiprocessing.Process.__init__(self)
        self.start_time = None

    def time_elapsed(self):
        return time.time() - self.start_time

class Script(object):
    """
    Tuples of times and actions.
    Actions are translated to key-presses that are sent to a VM,
    for example KeyPress.SHORTCUT_HOMEPAGE will resolve to 'alt+home'
    with in turn is the shortcut to go to the home(start)- page in many
    browsers.
    """
    
    def __init__(self, script):
        """
        Initialize a script.
        @param script: List of tuples, e.g. [(5, ''), (10,'')] .
        """
        # TODO It is assumed that the script times are ordered. Fix that.
        self.script = script

    def needs_action(self, time):
        """
        @return: True if an action should be done at given time or an action
                 should have been done but hasn't yet.
        """
        if self.script is None:
            return False
        if len(self.script) == 0:
            return False
        (at_time,__action) = self.script[0]
        if at_time <= time:
            return True
        return False

    def next_action(self):
        """Next action."""
        
        try:
            (__t,action) = self.script[0]
            return action
        except:
            return None

    def action_done(self):
        """Mark action as done (remove first list item)."""

        try:
            del(self.script[0])
        except:
            pass

class KeyPress(object):
    """Key-presses to use with KVMs sendkeys-command."""

    SHORTCUT_HOMEPAGE = 'alt-home'
    SHORTCUT_CLOSE_WINDOW = 'alt-f4'


class KVMHCRunProcess(multiprocessing.Process):
    """
    Process that starts a vm in a subprocess and talks to it via stdin.
    """

    SLEEP_STEP = 1

    def __init__(self, hc, script, timeout, command_queue, result_queue, work_dir):
        """
        Constructor.
        @param hc: An HCControl object.
        @param script: A script to be executed (list of pairs (second,action)).
        @param timeout: When the subprocesses will be terminated, in seconds.
        @param command_queue: Regularly polled by process, if non-empty terminate.
        @param result_queue: When process is finished, but path to the dump file.
        @param work_dir: Where the dump file will be placed.
        """

        multiprocessing.Process.__init__(self)
        
        self.hc = hc
        self.script = Script(script)
        self.cp_img_path = ''
        self.snapshot_path = ''
        self.vnc_port = None
        self.start_time = None
        self.dump_file = ''
        self.kvm_subprocess = None
        self.timeout = timeout
        self.daemon = True
        self.work_dir = work_dir
        self.__prepare_kvm_files()
        self.result_queue = result_queue
        self.command_queue = command_queue

    def __del(self):
        """Destructor."""

        # TODO ensure files (from __prepare_kvm_files) are removed.
        pass

    def __prepare_kvm_files(self):
        """Prepare files and paths (e.g. copy image file)."""

        if not path.isdir(self.work_dir):
            makedirs(self.work_dir)

        _handle_d, self.dump_file = tempfile.mkstemp(
                                            prefix=path.basename('%s_' %
                                                self.hc.image_snapshot),
                                            suffix='.dump',
                                            dir=self.work_dir)

        self.cp_img_path = self.hc.image_file
        self.snapshot_path = self.hc.image_snapshot
        # TODO Ensure secure chmod.

    def __kvm_cmd(self):
        """Returns list used to open kvm command subprocess."""

        usermode_net_params = ''
        if self.hc.own_ip:
            usermode_net_params = ',net=%s' % self.hc.own_ip
        if self.hc.dns_ip:
            if not usermode_net_params.endswith(','):
                usermode_net_params += ','
            usermode_net_params += 'dns=%s' % self.hc.dns_ip
        if self.hc.gateway_ip:
            if not usermode_net_params.endswith(','):
                usermode_net_params += ','
            usermode_net_params += 'host=%s' % self.hc.gateway_ip
             
        migration_cmd = ''
        # Snapshot is compressed?
        if self.snapshot_path.endswith('gz'):
            migration_cmd = 'exec:gzip -dc  %s' %  self.snapshot_path
        else:
            migration_cmd = 'exec:cat %s' % self.snapshot_path

        # TODO Ensure proper escaping, research and document here
        #      what really happens.
        # Maybe kvm should be started as --daemonize, or without nowait
        # for the monitor socket.
        cmd = [ "/usr/bin/kvm",
                "-hda", self.cp_img_path,
                "-net", "nic", # Create Network Interface Card in VLAN 0
                "-net", "user%s" % usermode_net_params, # Use user mode network
                "-net", "dump,file=%s" % self.dump_file, # Dump
                '-incoming', migration_cmd, # VM State to load.
                '-snapshot', # Do not alter files.
                '-m', self.hc.image_memory
               ]
        if self.vnc_port != None:
             cmd += [ "-vnc",  ":%d" % self.vnc_port ]
        else:
             #cmd.append('-nographic')
             cmd += [ "-vnc",  "none" ]
        cmd += [ "-monitor", "stdio"]

        return cmd

    def run(self):
        """Invoked on start() - call.
           Starts a KVM subprocess and a parallell interaction process which
           sends keyboard-press commands to the kvms monitor socket.
           After timeout, abort the process.
        """

        cmd = self.__kvm_cmd()

        logger.debug("Starting HC with: %s" % cmd)

        # Start kvm subprocess.
        # TODO Drop privileges as much as possible
        self.kvm_subprocess = subprocess.Popen(cmd, stdin=subprocess.PIPE,
                                               stdout=subprocess.PIPE,
                                               stderr=subprocess.PIPE,
                                               close_fds=True)

        # Init time.
        self.start_time = time.time()

        while self.elapsed_time() < self.timeout:
            self.kvm_subprocess.poll()
            if self.kvm_subprocess.returncode is not None:
                # TODO escalate this
                stdout_value, stderr_value = self.kvm_subprocess.communicate()
                logger.warning("KVM Process seemed to have died all by itself.")
                logger.warning("stderr: %s" % stderr_value)
                logger.warning("stdout: %s" % stdout_value)
                logger.warning("exit code: %d" % self.kvm_subprocess.returncode)
                break

            try:
                cmd = self.command_queue.get_nowait()
                # If command is given, this is the command to stop.
                break
            except Empty:
                pass

            while self.script.needs_action(self.elapsed_time()):
                action = self.script.next_action()
                self.kvm_subprocess.stdin.write('sendkey %s\n\r' % action)
                self.kvm_subprocess.stdin.flush()
                self.script.action_done()
                logger.debug("Sent keypress (%s) to VM." % action)

            sleep(KVMHCRunProcess.SLEEP_STEP)

        # Terminate the Processes
        if self.__terminate(self.kvm_subprocess) == None:
            # TODO Ensure that the process is joined.
            pass
        self.cleanup()
        self.result_queue.put(self.dump_file)

    def __terminate(self, process):
        """ Terminates the kvm subprocess and waits for it to
            shut down.  Blocks for up to 10 seconds before
            sending sigkill.

            Returns the subprocesses return code
        """

        state = process.poll()
        # TODO Send quit command to process.
        #process.stdin.write('quit\n\r')
        if state == None:
            process.terminate()
        waited = 0
        while state == None and waited < 10:
            state = process.poll()
            if state:
                break
            sleep(1)
            waited += 1
        if state == None:
            logger.warning("KVM subprocess not terminated after 10 seconds.")
            process.kill()
        return state

    def graceful_stop(self):
        """Stop remaining (kvm) processes."""

        logger.warning("Gracefull stopping not yet implemented (use command queue in future).")

    def stop_processes(self):
        """Big red button."""
        # TODO examine issue why this doesnt work
        # (probably related to process/forking)

        if self.kvm_subprocess is not None:
            self.kvm_subprocess.terminate()
        if self.kvm_subprocess is not None:
            self.kvm_subprocess.kill()
            #join()?

    def cleanup(self):
        """Remove socket and copied image file (but not dump)."""

        logger.debug('Cleaning up kvm process, leaving dump file.')

    def elapsed_time(self):
        """Calculates the elapsed time since start."""
        # TODO candidate for superclass TimedProcess
        return time.time() - self.start_time

    def time_to_live(self):
        """Calculates the time to live."""
        # TODO candidate for superclass TimedProcess
        return self.timeout - self.elapsed_time()

