# -*- coding: utf-8 -*-

# mapWOC
#
# Description: Node- centric classes.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
#
# Copyright:
# Copyright (C) 2011 Bundesamt für Sicherheit in der Informationstechni
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import ConfigParser
import logging
import multiprocessing
from os import path, listdir, remove, makedirs
from Queue import Empty
from shutil import copy
import socket
import subprocess
import time
import tempfile

from kvmhcrunprocess import KVMHCRunProcess, KeyPress
from analyserprocess import AnalysisProcess

from mapwoc.lib.hc import HC
from mapwoc.lib.mapwocqueue import Mapwocqueue
from mapwoc.lib.hccontrol import HCControl
from mapwoc.lib.scanconfiguration import VisitorConfiguration, VisitorBundle
from mapwoc.lib.result import Result

logger = logging.getLogger("node")

class NodeState(object):
    """Representing a state of a Node as communicated to other components."""

    IDLE,RUNNING,MANUAL,ANALYSING,STOPPING,ERROR = range(1,7)
    strstate = {IDLE:'idle',
                RUNNING:'running',
                MANUAL:'manual',
                ANALYSING:'analysis',
                STOPPING:'stopping',
                ERROR:'error'}

    def __init__(self):
        """Inits state to IDLE, empty runlist."""

        self.state = self.IDLE
        self.runlist = []

    def set(self, new_state):
        """Sets state to new state."""

        self.state = new_state

    def export(self):
        """Returns a dictionary describing the state as expected by the master
           module.
        """
        return {'' : self.strstate[self.state],
                'runlist' : self.runlist }

class Node:
    """The NodeCtrl."""

    VERSION = "0.1.0"

    def __init__ (self, configfile):
        """Initializes an empty Node, read in config and hc config."""

        self.__hcs = [] # Available honeyclients, HCControls
        self.__hc_instances = [] # Running (kvm) child processes

        # State
        self.state = NodeState()

        self.__read_conf(configfile)
        self.__load_hcs()
        self.next_vnc_port = 21 # For debuggin purposes, vnc ports
        self.visitor_configuration = None # Part of the scanconfiguration
        self.__kill_pill = False # Marker whether we should gracefully stop.
        self.__analyser_job_queue = Mapwocqueue()
        self.__analyser_result_queue = Mapwocqueue()

        # Start as many analyser-processes as specified in the config file.
        for i in range(0, self.__num_analyser_processes):
            self.__analyser_instances = [AnalysisProcess(self.__analyser_job_queue,
                                                 self.__analyser_result_queue,
                                                 self.__work_dir,
                                                 self.__result_dir,
                                                 self.__analyser_program)]
        for analyser in self.__analyser_instances:
            analyser.start()

        if not path.isdir(self.__work_dir):
            makedirs(self.__work_dir)
        if not path.isdir(self.__result_dir):
            makedirs(self.__result_dir)

        # TODO Clean up directories and files:
        for _file in listdir(self.__work_dir):
            logger.warning("Remaining %s in working directory (%s): ignoring." % (_file,
                path.abspath(self.__work_dir)))

        for _file in listdir(self.__result_dir):
            logger.warning("Remaining %s in result directory (%s): ignoring." % (_file,
                path.abspath(self.__result_dir)))

    def get_hcs(self):
        """Accessor to hcs (HCControl)."""

        return self.__hcs

    def get_hc(self, hcsid):
        """Accessor for a HCControl with given hcsid.
           Returns the HCControl, or None if not found.
        """

        for h in self.__hcs:
            if h.hcsid == hcsid:
                return h

        return None

    def start_hc(self, hc, visitorconfig):
        """Starts a KVM HC and returns its process."""

        kvm_process = KVMHCRunProcess(hc, [(5, KeyPress.SHORTCUT_HOMEPAGE)],
                                      visitorconfig.active_window,
                                      multiprocessing.Queue(),
                                      multiprocessing.Queue(), self.__work_dir)
        kvm_process.vnc_port = self.next_vnc_port
        self.next_vnc_port += 1
            
        # Fire.
        kvm_process.start()
        return kvm_process

    def stop_all(self):
        """Kill all processes, the hard way."""

        self.state.set(NodeState.STOPPING)
        while len(self.__hc_instances) != 0:
            instance = self.__hc_instances.pop()
            instance.stop_processes()
            instance.terminate()
            # join?

            # We now have dumps that should be saved and analysed somewhere.
            # Create an analyser instance / put in analysers job queue
            #copy(instance.copied_img_path + "dump", workingdirectory/dump/unanalysed"

            # TODO use a instance.cleanup()
            remove(instance.dump_file)
            remove(instance.qemu_monitor_socket)
            remove(instance.cp_img_path)
        self.visitor_configuration = None
        self.state.set(NodeState.ANALYSING)
        self.state.runlist = []

    def job_feasible(self, visitorconfig):
        """Checks whether the given job is feasible for this node.
           This currently means that the sum of ram requirements of the given
           HCs is less than the configured maxram in the configuration file.
        """

        ram_used = 0
        for vb in visitorconfig.visitors:
            hc = self.get_hc(vb.hcsid)
            if hc == None:
                logger.warning("While checking jobs feasibility, could not"
                               " find hc %s." % vb.hcsid)
                continue
  
            ram_used += int(hc.image_memory) * int(vb.number)
        logger.debug("Checking feasibility of job: takes %d, have %d ram" % (ram_used, self.__ressource_mb_ram))
        return ram_used <= self.__ressource_mb_ram

    def start_job(self, visitorconfig):
        """Starts a job, running given numbers of given hcsid
           with given livespan until stop is called.
        """

        # If already scanning, cannot start a new one.
        if self.visitor_configuration:
            logger.debug("Cannot start new scan: already scanning.")
            return False

        self.visitor_configuration = visitorconfig
        if visitorconfig is None:
            logger.debug("Cannot start new scan: No VisitorConfiguration given.")
            return False
        self.__kill_pill = False
        for vb in self.visitor_configuration.visitors:
            logger.info("Starting job with hcsid: %s" % vb.hcsid)
            hc = self.get_hc(vb.hcsid)
            if hc == None:
                logger.warning("Could not find HC (%s), part of a job, to start." % vb.hcsid)
                continue
  
            for i in range(vb.number):
                try:
                    kvm_process = self.start_hc(hc, visitorconfig)
                    self.__hc_instances += [kvm_process]

                    self.state.runlist.append(hc.hcsid)
                    self.state.set(NodeState.RUNNING)
                except:
                    # Most likely kvm_process couldnt start the machine or could not copy the files etc.
                    logger.warning("Error while starting the kvm-process.")
                    # TODO Cleanup
                    return False

        return True

    def stop_run(self, graceful=False):
        """Stops a current running job."""

        # Remove all conf, cleanup, set state, etc.
        logger.info("Will stop run.")
        if self.state.state is not NodeState.IDLE:
            self.state.set(NodeState.STOPPING)
        
        self.__kill_pill = True
        if not graceful:
            for process in self.__hc_instances:
                process.command_queue.put('stop')
            # TODO cleanup
        self.visitor_configuration = None

    def get_result_dir(self):
        """Returns the location where the results are stored"""
        return path.abspath(self.__result_dir)

    def __read_conf(self, config_file):
        """Read the config file."""

        config = ConfigParser.RawConfigParser()
        config.read(config_file)
        self.__hcsdir = config.get('config', 'hcsdir')
        self.__work_dir = config.get('config', 'work_dir')
        self.__result_dir = config.get('config', 'result_dir')
        self.__num_analyser_processes = config.getint('config', 'num_analyser')
        self.__ressource_mb_ram = 1024 * config.getfloat('config', 'max_gb_ram')
        self.__analyser_program = config.get('config', 'analyser')

    def __load_hcs(self):
        """Add HCs for which a config file is found in config path."""

        if not path.exists(self.__hcsdir):
            logger.warning("The HC directory %s does not exist." % self.__hcsdir)
            return

        for rel_hc_file in listdir(self.__hcsdir):
            abs_hc_file = path.normpath(path.join(self.__hcsdir, rel_hc_file))
            new_hc = HCControl(abs_hc_file)
            if len([hc for hc in self.__hcs if hc.hcsid == new_hc.hcsid]) > 0:
                logger.critical("Two configurations for hcsid %s "
                                "in HC directory %s found, exiting." % 
                                (new_hc.hcsid, self.__hcsdir))
                exit(5)
            self.__hcs += [new_hc]


    def get_hcsids(self):
        """Returns list with hcsids of available (startable) honeyclients."""

        answer = [hc.hcsid for hc in self.__hcs]
        return answer
        
    def poll_hcs(self):
        """Look for results in queues to kvm processes, repopulate state.
           If kill_pill is not present (graceful stop), restart finished hcs.
        """

        # Iterate over processes, reverse order, and remove finished ones.
        for i in range(len(self.__hc_instances)-1, -1, -1):
            kvm_process = self.__hc_instances[i]
            try:
                # TODO Process might die without putting stuff in result queue.
                #      Check processes time to live and state, too.
                r = kvm_process.result_queue.get_nowait()
                # Tell it to clean up, store result, start analysis.
                logger.info("HC %s finished." % kvm_process.hc.hcsid)
                kvm_process.stop_processes()
                kvm_process.cleanup()
                self.__analyser_job_queue.put(kvm_process.dump_file)
                # TODO Instead of kill_pill, use state (gracefully stopping).
                if not self.__kill_pill:
                    # Create and start new process.
                    n_kvm_process = self.start_hc(kvm_process.hc,
                                                  self.visitor_configuration)
                    self.__hc_instances.append(n_kvm_process)
                    logger.info("Restarting a HC.")
                del self.__hc_instances[i]
            except Empty:
                logger.debug("One HC did not yet deliver a result.")
                if not kvm_process.is_alive():
                    logger.warning("panic: kvm process died!")
                    

        try:
            while not self.__analyser_result_queue.empty():
                result = self.__analyser_result_queue.get_nowait()
                self.pack_result(result)
                logger.warning("Have a result !")
        except Empty:
            logger.debug("No Analysis results.")

        self.__update_runlist()

    def pack_result(self, result):
        """Stores a result to be fetched by master, later."""
        
        # Need to pack also visitorconf (of that time, hc details, etc)
        # Make unique directory name.
        # dirname = tempfile.mkdtemp(prefix="%s_"%'result.hcsid', dir=self.__result_dir)
        logger.warning("Shall create the temporary result directory (which then shall be synced.).")


    def __update_runlist(self):
        """Updates runlist in state dict.
           Also, updates general state.
        """

        self.state.runlist = [hc.hc.hcsid for hc in self.__hc_instances]
        if len(self.__hc_instances) == 0 and self.visitor_configuration == None:
            # Set state to idle if no analysis running.
            if (self.__analyser_job_queue.empty()
                and self.__analyser_result_queue.empty()):
            # TODO: Find out why __analyser_resulte_queue.objects_put is always 0 here
            # if (self.__analyser_job_queue.objects_put == self.__analyser_result_queue.objects_put):
                self.state.set(NodeState.IDLE)
            else:
                self.state.set(NodeState.ANALYSING)

    def __print_hcs(self):
        """Prints available and running HCs to stdout."""

        print "Available HCs:"
        for h in self.__hcs:
            print "  " + h.hcsid + " (" + h.name + ")"
        print
        print "Running HCs:"
        for h in self.__hc_instances:
            print "  " + h.hcsid + " (" + h.name + ")"


